from storage import StorableObject
from quixote import get_request, get_session

import emails

from form import *
from misc import simplify

from formdata import FormData
from roles import Role
from categories import Category

widget_classes = {
    'string': StringWidget,
    'email': EmailWidget,
    'text': TextWidget,
    'bool': CheckboxWidget,
    'item': SingleSelectWidget,
    'file': FakeFileWidget,
}

widget_extra_attributes = {
    'text': ('cols', 'rows'),
    'string': ('size',),
}

status_labels = {
    None: N_('Unknown'),
    'new': N_('New'),
    'rejected': N_('Rejected'),
    'accepted': N_('Accepted'),
    'finished': N_('Finished')
}

class FormField:
    id = None
    label = None
    type = None
    required = True
    in_listing = True

    # for type == 'items'
    items = None
    show_as_radio = False

    # for type == 'text'
    cols = None
    rows = None
    pre = False

    # for type == 'string'
    size = None

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k.replace('-', '_'), v)


class FormDef(StorableObject):
    _names = 'formdefs'

    name = None
    url_name = None
    fields = None
    receiver_id = None
    category_id = None
    roles = None
    discussion = False
    confirmation = True
    public = False
    detailed_emails = False

    def migrate(self):
        changed = False

        if self.__dict__.has_key('receiver'):
            self.receiver_id = self.__dict__['receiver']
            del self.__dict__['receiver']
            changed = True

        if self.__dict__.has_key('category'):
            self.category_id = self.__dict__['category']
            del self.__dict__['category']
            changed = True

        if not self.url_name:
            try:
                int(self.id)
            except ValueError:
                self.url_name = self.id
            else:
                self.url_name = simplify(self.name)
            changed = True

        if self.fields and type(self.fields[0]) is dict:
            for f in self.fields:
                if f.has_key('name'):
                    f['label'] = f['name']
                    del f['name']
            self.fields = [FormField(**x) for x in self.fields]
            for i, f in enumerate(self.fields):
                f.id = str(i)
            for formdata in self.data_class().select():
                for f in self.fields:
                    if not formdata.data.has_key(f.label):
                        continue
                    formdata.data[f.id] = formdata.data[f.label]
                    del formdata.data[f.label]
                formdata.store()
            changed = True

        if changed:
            self.store()

    def data_class(self):
        cls = FormData
        cls._names = 'form-%s' % self.url_name
        return cls


    def get_receiver(self):
        if self.receiver_id:
            return Role.get(self.receiver_id)
        else:
            return None

    def set_receiver(self, role):
        if role:
            self.receiver_id = role.id
        elif self.receiver_id:
            self.receiver_id = None
    receiver = property(get_receiver, set_receiver)

    def get_category(self):
        if self.category_id:
            return Category.get(self.category_id)
        else:
            return None

    def set_category(self, category):
        if category:
            self.category_id = category.id
        elif self.category_id:
            self.category_id = None
    category = property(get_category, set_category)

    def get_by_urlname(cls, url_name):
        objects = [x for x in cls.select() if x.url_name == url_name]
        if objects:
            return objects[0]
        raise KeyError()
    get_by_urlname = classmethod(get_by_urlname)


    def create_form(self):  # XXX: merge create_form and create_view_form
        form = Form(enctype = "multipart/form-data", use_tokens = False) #, use_tokens = not self.confirmation)
        for field in self.fields:
            if field.type == 'title':
                form.widgets.append(HtmlWidget(htmltext('<h3>%s</h3>' % field.label)))
                continue
            if field.type == 'subtitle':
                form.widgets.append(HtmlWidget(htmltext('<h4>%s</h4>' % field.label)))
                continue
            if field.type == 'comment':
                form.widgets.append(HtmlWidget(htmltext('<p>%s</p>' % field.label)))
                continue
            
            widget_class = widget_classes[field.type]

            kwargs = {'required': field.required}
            if field.type == 'item':
                kwargs['options'] = field.items or [(None, '---')]
                if field.show_as_radio:
                    widget_class = RadiobuttonsWidget
            if widget_extra_attributes.has_key(field.type):
                for k in widget_extra_attributes[field.type]:
                    if hasattr(field, k):
                        kwargs[k] = getattr(field, k)
            form.add(widget_class, 'f%s' % field.id, title = field.label, **kwargs)
        return form

    def create_view_form(self, dict = {}):
        form = Form(enctype = "multipart/form-data")
        session = get_session()
        for field in self.fields:
            kwargs = {}
            if field.type == 'title':
                form.widgets.append(HtmlWidget(htmltext('<h3>%s</h3>' % field.label)))
                continue
            if field.type == 'subtitle':
                form.widgets.append(HtmlWidget(htmltext('<h4>%s</h4>' % field.label)))
                continue
            if field.type == 'comment':
                form.widgets.append(HtmlWidget(htmltext('<p>%s</p>' % field.label)))
                continue

            widget_class = widget_classes[field.type]
            value = dict.get(field.label, '')
            if field.type == 'item':
                widget_class = StringWidget

            if field.type == 'file':
                kwargs['preview'] = True
            if field.type == 'bool':
                kwargs['disabled'] = 'disabled'

            if widget_extra_attributes.has_key(field.type):
                for k in widget_extra_attributes[field.type]:
                    if hasattr(field, k):
                        kwargs[k] = getattr(field, k)

            if field.type == 'file':
                widget_class = FakeFileWidget

            form.add(widget_class, 'f%s' % field.id, title = field.label,
                    value = value, readonly = 'readonly', **kwargs)
        return form

    def get_data(self, form):
        d = {}
        for field in self.fields:
            widget = form.get_widget('f%s' % field.id)
            if widget:
                d[field.id] = widget.parse()
        return d

    def notify_new(self, formdata):
        if not self.receiver.emails:
            return
        url = '%s/%s/status' % (get_request().get_url(1), formdata.id)
        mail_body = _("""Hi,

A new form has been submitted on the website, you can consult it with this
link: %(url)s
""") % {'url': url}
        if self.detailed_emails:
            details = []
            if formdata.user:
                details.append(_('User name'))
                details.append('  %s' % formdata.user.name)
            fields = self.fields
            for field in self.fields:
                data = formdata.data
                if not data.has_key(field.id):
                    continue
                details.append(field.label)
                if field.type == 'bool':
                    if data[field.id]:
                        details.append('  %s' % _('Yes'))
                    else:
                        details.append('  %s' % _('No'))
                elif field.type == 'file':
                    file_url = url.replace('status', 'download?f=%d' % field.id)
                    details.append('  %s' % file_url)
                elif field.type == 'text':
                    # XXX: howto support preformatted text in a dl in docutils ?
                    details.append(('  %s' % data[field.id]).replace('\n', '\n  '))
                else:
                    details.append('  %s' % data[field.id])
            mail_body += '\n\n' + '\n'.join(details)

        emails.email(_("New form (%s)") % self.name, mail_body,
                self.receiver.emails[0],
                bcc = self.receiver.emails[1:])

    def notify_change_user(self, formdata, old_status):
        if not formdata.user:
            return
        if not formdata.user.email:
            return

        if self.receiver.emails:
            from_email = self.receiver.emails[0]
        else:
            from_email = None
        
        url = '%s/' % get_request().get_url(1)
        if old_status != formdata.status:
            mail_body = _("""Hi,

Status of the form you submitted just changed (from "%(before)s" to "%(after)s"),
you can consult it with this link:

%(url)s
""") % {'url': url,
            'before': _(status_labels[old_status]),
            'after': _(status_labels[formdata.status])}
        else:
            mail_body = _("""Hi,

New information has been submitted on the form you submitted, you can
consult it with this link:

%(url)s
""") % {'url': url}

        mail_body += self.get_detailed_evolution(formdata)

        emails.email(_('Form status change'), mail_body, formdata.user.email,
                replyto = from_email)

    def get_detailed_evolution(self, formdata):
        if not self.detailed_emails or not formdata.evolution:
            return ''

        details = []
        evo = formdata.evolution[-1]
        if evo.who:
            details.append(_('User name'))
            details.append('  %s' % evo.who.name)
        if evo.status:
            details.append(_('Status'))
            details.append('  %s' % _(status_labels[evo.status]))
        if evo.comment:
            details.append('\n%s\n' % evo.comment)
        return '\n\n----\n\n' + '\n'.join(details)


    def notify_change_receiver(self, formdata):
        if not self.receiver.emails:
            return

        url = '%s/status' % get_request().get_url(1)
        mail_body = _("""Hi,

A form just changed, you can consult it with this link:

%(url)s
""") % {'url': url}
        mail_body += self.get_detailed_evolution(formdata)
            
        emails.email(_('Form status change (%s)') % self.name, mail_body,
                self.receiver.emails)

